@@ -170,10 +170,8 @@ setupexcludes(void)
 		excludeinsert(&address, bits, &rndexclqueue);
 	}
 
-	if ((stream = fopen(excludefile, "r")) == NULL) {
-		warn("fopen: %s", excludefile);
+	if ((stream = fopen(excludefile, "r")) == NULL)
 		return (-1);
-	}
 
 	while (fgets(line, sizeof(line), stream) != NULL) {
 		len = strlen(line);
