@@ -41,6 +41,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <pcap.h>
+#include <errno.h>
 #include <err.h>
 
 #include "config.h"
@@ -231,7 +232,12 @@ synprobe_send(struct in_addr dst, u_int32_t *seqnr)
 	current = (current + 1) % MAXSOCKS;
 	if (sendto(synsock[current], buf, sizeof(buf), 0,
 		   (struct sockaddr *)&sa, sizeof(sa)) == -1) {
-		warn("sendto");
+#ifdef EHOSTDOWN
+		if (errno == EHOSTDOWN)
+			return (0);
+#endif
+		if (errno != ENOBUFS)
+			warn("sendto");
 		return (-1);
 	}
 
