@@ -1,5 +1,5 @@
 /*
- * ScanSSH - simple SSH version scanner V1.4
+ * ScanSSH - simple SSH version scanner V1.5
  *
  * Copyright 2000-2001 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
@@ -132,18 +132,27 @@ int
 synlist_insert(struct argument *arg)
 {
 	struct timeval tmp;
+	struct argument *listarg;
 
 	if (!synlist_hasspace())
 		return (-1);
 
 	timerclear(&tmp);
-	tmp.tv_sec += SYNWAIT;
-	if (arg->a_retry > SYNRETRIES/2)
-		tmp.tv_sec += SYNWAIT;
+	tmp.tv_sec += (arg->a_retry/2 + 1) * SYNWAIT;
 
 	gettimeofday(&arg->a_tv, NULL);
 	timeradd(&tmp, &arg->a_tv, &arg->a_tv);
-	TAILQ_INSERT_TAIL(&synqueue, arg, a_next);
+
+	/* Insert in time order */
+	TAILQ_FOREACH_REVERSE(listarg, &synqueue, a_next, syn_list) {
+		if (timercmp(&listarg->a_tv, &arg->a_tv, <=))
+			break;
+	}
+	if (listarg)
+		TAILQ_INSERT_AFTER(&synqueue, listarg, arg, a_next);
+	else
+		TAILQ_INSERT_TAIL(&synqueue, arg, a_next);
+
 	synqueuesz++;
 
 	return (0);
@@ -222,8 +231,8 @@ scanips(struct argument *args, u_int32_t nargs)
 	u_int32_t entries = 0;
 	fd_set *readset, *writeset;
 	int maxfd = 0, fdsetsz, counts = 0;
-	struct timeval tv, ptv;
-	int res, i;
+	struct timeval tv;
+	int res, i, lastinsert, newinsert;
 	char buf[MAXBUF], *p;
 
 	for (i = 0; i < MAX_PROCESSES; i++) {
@@ -240,7 +249,10 @@ scanips(struct argument *args, u_int32_t nargs)
 
 	TAILQ_INIT(&readyqueue);
 	synlist_init();
-	timerclear(&ptv);
+
+	/* Keep track where which process received the last insert */
+	lastinsert = 0;
+
 	while (TAILQ_FIRST(&genqueue) || entries ||
 	       counts || !synlist_empty()) {
 		/* Create new entries, if we need them */
@@ -256,8 +268,25 @@ scanips(struct argument *args, u_int32_t nargs)
 			if (results[i] >= 0)
 				FD_SET(results[i], readset);
 		}
+
+		/* Find new timeout */
 		timerclear(&tv);
-		tv.tv_usec = 500000;
+		tv.tv_usec = 250000;
+		if (!synlist_empty()) {
+			struct timeval ctv, tmp;
+			struct argument *sarg = TAILQ_FIRST(&synqueue);
+
+			gettimeofday(&ctv, NULL);
+			tmp = sarg->a_tv;
+			/* If it overflows, it will become bigger */
+			timersub(&tmp, &ctv, &tmp);
+			if (timercmp(&tmp, &tv, <))
+				tv = tmp;
+			timerclear(&tmp);
+			if (timercmp(&tmp, &tv, >))
+				tv = tmp;
+		}
+
 		/* Wait in select until there is a connection. */
 		if ((res = select(maxfd + 1, readset, writeset, NULL,
 				  &tv)) == -1) {
@@ -307,7 +336,7 @@ scanips(struct argument *args, u_int32_t nargs)
 		}
 
 		if (usepcap) {
-			int res, nsent;
+			int res, nsent, probefailed;
 			struct in_addr address;
 			struct argument *addrchk;
 			struct timeval now;
@@ -332,34 +361,39 @@ scanips(struct argument *args, u_int32_t nargs)
 
 			gettimeofday(&now, NULL);
 			nsent = 0;
-			if (timercmp(&now, &ptv, >)) {
-				/* Remove all expired entries */
-				while (synlist_expired() && nsent < MAXBURST) {
-					struct argument *exp;
-
-					exp = synlist_dequeue();
-					if (exp->a_retry < SYNRETRIES) {
-						exp->a_retry++;
-						if (synlist_probe(exp) == -1)
-							break;
-						nsent++;
-					} else
-						printres(exp, "<timeout>");
-				}
-				while (entries && nsent < MAXBURST &&
-				       synlist_hasspace()) {
-					entries--;
-					args[entries].a_retry = 0;
-					if (synlist_probe(&args[entries])== -1)
+			probefailed = 0;
+			/* Remove all expired entries */
+			while (synlist_expired() && nsent < MAXBURST) {
+				struct argument *exp;
+				
+				exp = synlist_dequeue();
+				if (exp->a_retry < SYNRETRIES) {
+					exp->a_retry++;
+					if (synlist_probe(exp) == -1) {
+						probefailed = 1;
 						break;
+					}
 					nsent++;
-				}
-				if (nsent == MAXBURST) {
-					timerclear(&ptv);
-					ptv.tv_usec = 300000;
-					timeradd(&now, &ptv, &ptv);
-				}
+				} else
+					printres(exp, "<timeout>");
 			}
+
+			/* Put new entries in the synlist */
+			while (!probefailed && entries && nsent < MAXBURST &&
+			       synlist_hasspace()) {
+				entries--;
+				args[entries].a_retry = 0;
+				if (synlist_probe(&args[entries])== -1) {
+					probefailed = 1;
+					break;
+				}
+				nsent++;
+			}
+#ifdef DEBUG
+			fprintf(stderr, "Sent: %d %s (%d-%d:%d)\n", nsent,
+				probefailed ? "failed" : "",
+				synqueuesz, entries, counts);
+#endif /* DEBUG */
 		} else {
 			/* No advanced syn probing, place data direcly
 			 * on ready queue.
@@ -371,7 +405,12 @@ scanips(struct argument *args, u_int32_t nargs)
 			}
 		}
 
-		for (i = 0; i < MAX_PROCESSES; i++) {
+		/* Round robin the inserts */
+		i = (lastinsert + 1) % MAX_PROCESSES;
+		newinsert = lastinsert;
+		while (i != lastinsert) {
+			i = (i + 1) % MAX_PROCESSES;
+
 			if (commands[i] < 0 ||
 			    !FD_ISSET(commands[i], writeset))
 				continue;
@@ -382,6 +421,7 @@ scanips(struct argument *args, u_int32_t nargs)
 				atomicio(write, commands[i], arg,
 					 sizeof(struct argument));
 				counts++;
+				newinsert = i;
 			} else if (synlist_empty()) {
 				struct argument arg;
 				memset(&arg, 0, sizeof(arg));
@@ -391,6 +431,7 @@ scanips(struct argument *args, u_int32_t nargs)
 				commands[i] = -1;
 			}
 		}
+		lastinsert = newinsert;
 	}
 }
 
@@ -443,9 +484,9 @@ usage(char *name)
 {
 	fprintf(stderr, 
 		"%s: [-ERh] [-e excludefile] [-p ifaddr] <IP address|network>...\n\n"
-		"\t-E          do not exit if exclude file is missing,\n"
+		"\t-E          exit if exclude file is missing,\n"
 		"\t-R          do not honor exclude file for random addresses,\n"
-		"\t-e <file>   exlude the IP addresses and networks in <file>,\n"
+		"\t-e <file>   exclude the IP addresses and networks in <file>,\n"
 		"\t-p <ifaddr> specifies the local interface address,\n"
 		"\t-h          this message.\n\n",
 		name);
@@ -594,8 +635,12 @@ next_address(struct generate *gen, struct in_addr *addr)
 		if (!random || rndexclude) {
 			tmp = exclude(ipv4addr, &excludequeue);
 			if (ipv4addr.s_addr != tmp.s_addr) {
-				if (random)
+				if (random) {
+					if (gen->gen_flags & FLAGS_SUBTRACTEXCLUDE)
+						gen->gen_max--;
+
 					continue;
+				}
 
 				/* In linear mode, we can skip these */
 				offset = gen->gen_current;
@@ -627,16 +672,19 @@ next_address(struct generate *gen, struct in_addr *addr)
 			}
 		}
 		
-		if (gen->gen_flags & FLAGS_USERANDOM) {
+		if (random) {
 			tmp = exclude(ipv4addr, &rndexclqueue);
 			if (ipv4addr.s_addr != tmp.s_addr) {
+				if (gen->gen_flags & FLAGS_SUBTRACTEXCLUDE)
+					gen->gen_max--;
 				continue;
 			}
 		}
 		
 		/* We have an address */
 		done = 1;
-	} while ((gen->gen_current < gen->gen_end) && !done);
+	} while ((gen->gen_current < gen->gen_end) && 
+		 (gen->gen_n < gen->gen_max) && !done);
 
 	if (!done)
 		return (-1);
@@ -684,7 +732,8 @@ address_node_get(char *line)
  * [split(x/n)/][random(x,s)/][(]<address/mask> .... [)]
  */
 
-int generate_split(struct generate *gen, char **pline)
+int
+generate_split(struct generate *gen, char **pline)
 {
 	char *line, *end;
 
@@ -720,7 +769,8 @@ int generate_split(struct generate *gen, char **pline)
  * [split(x/n)/][random(x,s)/][(]<address/mask> .... [)]
  */
 
-int generate_random(struct generate *gen, char **pline)
+int
+generate_random(struct generate *gen, char **pline)
 {
 	int i;
 	char seed[31], *line, *end;
@@ -759,6 +809,13 @@ int generate_random(struct generate *gen, char **pline)
 
 	gen->gen_flags |= FLAGS_USERANDOM;
 
+	/* If the random numbers exhaust all possible addresses,
+	 * we need to subtract those addresses from the count
+	 * that can not be generated because they were excluded
+	 */
+	if (!gen->gen_max)
+		gen->gen_flags |= FLAGS_SUBTRACTEXCLUDE;
+
 	return (0);
 }
 
@@ -887,7 +944,7 @@ main(int argc, char **argv)
 	int ch;
 	struct argument *args = NULL;
 	u_int32_t nargs;
-	int failonexclude = 1;
+	int failonexclude = 0;
 
 	name = argv[0];
 	while ((ch = getopt(argc, argv, "hp:e:ER")) != -1)
@@ -897,9 +954,9 @@ main(int argc, char **argv)
 			break;
 		case 'e':
 			excludefile = optarg;
-			break;
+			/* FALLTHROUGH */
 		case 'E':
-			failonexclude=0;
+			failonexclude = 1;
 			break;
 		case 'R':
 			rndexclude=0;
@@ -931,8 +988,10 @@ main(int argc, char **argv)
 
 	setupchildren();
 
-	if (setupexcludes() == -1 && failonexclude)
+	if (setupexcludes() == -1 && failonexclude) {
+		warn("fopen: %s", excludefile);
 		exit(1);
+	}
 
 	nargs = EXPANDEDARGS;
 	args = xmalloc(nargs * sizeof(struct argument));
@@ -947,19 +1006,7 @@ main(int argc, char **argv)
 		argv++;
 		argc--;
 	}
-	/*
-	{
-		int i, n;
-		struct argument args[100];
 
-		while ((n = populate(args, 100)) != 0) {
-			for (i = 0; i < n; i++)
-				printres(&args[i], "good");
-		}
-
-		exit(1);
-	}
-	*/
 	if (!TAILQ_FIRST(&genqueue))
 		errx(1, "nothing to scan");
 
